# -*- coding: utf-8 -*-

import numpy as np

from sklearn.decomposition import PCA
from sklearn.neighbors import NearestNeighbors


def initialize_Z(X, n_neighbors, n_pcs, landmarks, random_state):
    
    m, n = X.shape
    if n_pcs is None:
        n_pcs = min(max(m // 50, 20), n-1)
    if n_pcs < m:
        X_new = PCA(n_components=n_pcs, svd_solver='arpack', random_state=random_state).fit_transform(X.T)
    else:
        n_pcs = None
        X_new = X.T
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, metric='euclidean').fit(X_new[landmarks, :])
    Zi = np.zeros((n, n))
    Zi[:, landmarks] = nbrs.kneighbors_graph(X_new).toarray()
    np.fill_diagonal(Zi, 0)
    
    return Zi, n_pcs






















# -*- coding: utf-8 -*-

from typing import Callable, Optional, Union
from anndata import AnnData
from ._compat import Literal

import torch
import numpy as np

from scipy.sparse import issparse, spmatrix, csr_matrix

from ._initialize import initialize_Z
from ._solve import solve_P, solve_Z


_Metric = Literal['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan',
                  'braycurtis', 'canberra', 'chebyshev', 'correlation',
                  'dice', 'hamming', 'jaccard', 'kulsinski', 'mahalanobis',
                  'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean',
                  'sokalmichener', 'sokalsneath', 'sqeuclidean', 'yule']
_Metric_fn = Callable[[np.ndarray, np.ndarray], float]


def subspace_analysis(
    adata: AnnData,
    n_neighbors: Optional[int] = None,
    n_pcs: Optional[int] = None,
    metric: Union[_Metric, _Metric_fn] = 'euclidean',
    Lambda: float = 0.1,
    n_iterations: int = 500,
    n_discriminant: Optional[int] = None,
    Z_mask: Optional[Union[np.ndarray, spmatrix]] = None,
    use_landmarks: Optional[bool] = None,
    use_highly_variable: Optional[bool] = None,
    device: Optional[str] = None,
    random_state: int = 0,
    key_added: Optional[str] = None,
    copy: bool = False,
) -> Optional[AnnData]:
    '''
    Subspace analysis for representation learning [Wong17]_ [Shi21]_.
    
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_neighbors
        Number of neighbors for constructing a :class:`~sklearn.neighbors.NearestNeighbors`
        graph as hard constraints in subspace analysis.
        Defaults to the number of samples devided by 10 with a minimum number of 20.
    n_pcs
        Number of principal components to use for constructing a
        :class:`~sklearn.neighbors.NearestNeighbors` graph as hard
        constraints in subspace analysis.
        Defaults to the number of features devided by 50 with a minimum number of 20.
    metric
        Distance metric to use for constructing a
        :class:`~sklearn.neighbors.NearestNeighbors` graph as hard
        constraints in subspace analysis.
    Lambda
        Hyperparameter for sparsity regularization.
    n_iterations
        Number of iterations for the optimization.
    n_discriminant
        Number of discriminant vectors to store for label transfer.
        By default stores all discriminant vectors.
    Z_mask
        Customized sample-by-sample graph of shape (`adata.n_obs`, `adata.n_obs`)
        as hard constraints in subspace analysis.
        By default computes a :class:`~sklearn.neighbors.NearestNeighbors` graph.
    use_landmarks
        Whether to use landmarks as hard constraints in subspace analysis,
        stored in `adata.obs['is_landmarks']`.
        By default uses them if they have been selected beforehand.
    use_highly_variable
        Whether to use highly variable genes only, stored in `adata.var['highly_variable']`.
        By default uses them if they have been determined beforehand.
    device
        The desired device for `PyTorch` computation. By default uses cuda if cuda is avaliable
        cpu otherwise.
    random_state
        Change to use different initial states for the optimization.
    key_added
        If not specified, the subspace data is stored in `adata.uns['subspace_analysis']`,
        representation is stored in `adata.obsp['representation']` and
        discriminant matrix is stored in `adata.varm['discriminant']`.
        If specified, the subspace data is added to `adata.uns[key_added]`,
        representation is stored in `adata.obsp[key_added+'_representation']` and
        discriminant matrix is stored in `adata.varm[key_added+'_discriminant']`.
    copy
        Return a copy instead of writing to ``adata``.
    
    Returns
    -------
    Depending on ``copy``, returns or updates ``adata`` with the following fields.
    
    See ``key_added`` parameter description for the storage path of
    representation and discriminant.
    
    representation : :class:`~scipy.sparse.csr_matrix` (.obsp)
        The subspace representation of samples.
    discriminant : :class:`~numpy.ndarray` (.uns[``key_added``])
        The discriminant vectors for label transfer.
    '''
    
    adata = adata.copy() if copy else adata
    
    if device is None:
        if torch.cuda.is_available():
            device = "cuda"
        else:
            device = "cpu"
    
    device = torch.device(device)
    
    if use_landmarks is None:
        use_landmarks = True if 'is_landmark' in adata.obs else False
    
    landmarks = np.flatnonzero(adata.obs['is_landmark']) if use_landmarks else np.arange(adata.n_obs)
    
    if n_neighbors is None:
        n_neighbors = min(max(adata.n_obs // 10, 20), adata.n_obs)
    
    if use_highly_variable is True and 'highly_variable' not in adata.var.keys():
        raise ValueError(
            'Did not find adata.var[\'highly_variable\']. '
            'Either your data already only consists of highly-variable genes '
            'or consider running `pp.highly_variable_genes` first.'
        )
    
    if use_highly_variable is None:
        use_highly_variable = True if 'highly_variable' in adata.var.keys() else False
    
    adata_use = (
        adata[:, adata.var['highly_variable']] if use_highly_variable else adata
    )
    
    if n_discriminant is None:
        n_discriminant = adata_use.n_vars
    
    
    X = adata_use.X.toarray().T if issparse(adata_use.X) else adata_use.X.T
    if Z_mask is None:
        Z_mask, n_pcs = initialize_Z(X, n_neighbors, n_pcs, landmarks, random_state)
    else:
        if len(Z_mask.shape) != 2 or Z_mask.shape[0] != X.shape[1] or Z_mask.shape[1] != X.shape[1]:
            raise ValueError(
                'The shape of Z_mask needs to be (adata.n_obs, adata.n_obs) '
                f'({adata.n_obs}, {adata.n_obs}), but given {Z_mask.shape}.'
            )
        Z_mask = Z_mask.toarray() if issparse(Z_mask) else Z_mask
        Z_mask[Z_mask != 0] = 1
        n_pcs = None
    
    X = torch.Tensor(X).type(torch.float32).to(device)
    Z_mask = torch.Tensor(Z_mask).type(torch.float32).to(device)
    
    P = torch.eye(adata_use.n_vars).to(device)
    
    X_input = torch.matmul(P.T, X)
    Z, E = solve_Z(X_input, Lambda, Z_mask, n_iterations, device)  
    
    P = solve_P(X, Z, P)
    
    Z = Z.cpu().numpy()
    P = P.cpu().numpy()
    
    eps = 1e-14
    Z[Z < eps] = 0
    
    
    if key_added is None:
        key_added = 'subspace_analysis'
        conns_key = 'representation'
        dists_key = 'representation'
    else:
        conns_key = key_added + '_representation'
        dists_key = key_added + '_representation'
    
    adata.uns[key_added] = {}
    
    subspace_dict = adata.uns[key_added]
    
    subspace_dict['connectivities_key'] = conns_key
    subspace_dict['distances_key'] = dists_key
    subspace_dict['discriminant'] = P[:, :n_discriminant]
    subspace_dict['var_names_use'] = adata_use.var_names.to_numpy()
    
    subspace_dict['params'] = {}
    subspace_dict['params']['n_neighbors'] = np.count_nonzero(Z) // Z.shape[0]
    subspace_dict['params']['n_pcs'] = n_pcs
    subspace_dict['params']['metric'] = str(metric)
    subspace_dict['params']['Lambda'] = Lambda
    subspace_dict['params']['n_iterations'] = n_iterations
    subspace_dict['params']['n_discriminant'] = n_discriminant
    subspace_dict['params']['use_landmarks'] = use_landmarks
    subspace_dict['params']['use_highly_variable'] = use_highly_variable
    subspace_dict['params']['random_state'] = random_state
    subspace_dict['params']['method'] = 'umap'
    
    adata.obsp[conns_key] = csr_matrix(Z)
    
    return adata if copy else None






















# -*- coding: utf-8 -*-

import numpy as np
from numba import jit


@jit(nopython=True)
def cyclic_descent_loop(X, g, w, rand_idx, shrinkFactor, threshold):
    
    for j in rand_idx:
        wjold = w[j]
        
        gj = g + w[j] * X[:, j]
        wj = np.sum(X[:, j] * gj)
        
        w[j] = np.sign(wj) * max(np.abs(wj) - threshold, 0) / shrinkFactor[j]
        
        g = g - X[:, j] * (w[j] - wjold)
    
    return w, g


def cyclic_descent(X, r, w, rand_idx, active, norms, threshold, rng):
    
    num_rand = rand_idx.shape[0]
    
    if num_rand == 0:
        return w, r
    else:
        rand_idx = rand_idx[rng.permutation(num_rand)]
    
    w, r = cyclic_descent_loop(X, r, w, rand_idx, norms, threshold)
    
    return w, r


def lasso_fit(diag_col, X, w, r, threshold, reltol, norms, rng):
    
    active = w != 0
    
    wold = w
    
    while True:
        
        rand_idx = np.setdiff1d(np.flatnonzero(active), diag_col)
        w, r = cyclic_descent(X, r, w, rand_idx, active, norms, threshold, rng)
        active = w != 0
        
        if np.linalg.norm((w - wold) / (1.0 + np.abs(wold)), np.inf) < reltol:
            
            wold = w
            
            potentially_active = np.abs(np.matmul(r, X)) > threshold
            
            if np.any(potentially_active):
                
                new_active = active | potentially_active
                
                rand_idx = np.setdiff1d(np.flatnonzero(new_active), diag_col)
                w, r = cyclic_descent(X, r, w, rand_idx, new_active, norms, threshold, rng)
                new_active = w != 0
            else:
                new_active = active
            
            if np.all(new_active == active):
                break
            else:
                active = new_active
            
            if np.linalg.norm((w - wold) / (1.0 + np.abs(wold)), np.inf) < reltol:
                break
        
        wold = w
        
    return w, r


def lasso_landmarks(X, S, stats, idx, rng):
    
    norms = stats['normsSt']
    XS = stats['XSt']
    
    w_size = S.shape[0]
    w_vec = stats['W'][:w_size, idx]
    
    diag_col = np.flatnonzero(S==idx)
    
    w_vec, r = lasso_fit(diag_col, XS, w_vec, stats['R'][:, idx],
                         stats['Lambda'], stats['reltol'], norms, rng)
    
    stats['R'][:, idx] = r
    stats['W'][:w_size, idx] = w_vec
    
    return stats






















# -*- coding: utf-8 -*-

from typing import Optional, Union
from anndata import AnnData

import numpy as np
import pandas as pd

from scipy.sparse import issparse
from sklearn.neighbors import KNeighborsClassifier
from natsort import natsorted

from ._subspace_analysis import _Metric, _Metric_fn


def transfer_labels(
    adata: AnnData,
    adata_learning: AnnData,
    groups: str,
    n_neighbors: int = 10,
    metric: Union[_Metric, _Metric_fn] = 'cosine',
    n_discriminant: Optional[int] = None,
    subspace_key: Optional[str] = None,
    copy: bool = False,
) -> Optional[AnnData]:
    '''
    Transfer labels from learning samples to other samples [Shi21]_.
    
    Parameters
    ----------
    adata
        Annotated data matrix.
    adata_learning
        Annotated data matrix of learning samples with given labels.
    groups
        The key of the observations grouping to transfer.
    n_neighbors
        Number of neighbors for :class:`~sklearn.neighbors.KNeighborsClassifier`.
    metric
        Distance metric for :class:`~sklearn.neighbors.KNeighborsClassifier`.
    n_discriminant
        Number of discriminant vectors to use.
        By default uses all stored discriminant vectors.
    subspace_key
        If not specified, it looks `adata_learning.uns['subspace_analysis']` for subspace settings
        and `adata_learning.varm['discriminant']` for discriminant matrix
        (default storage places for :func:`~lantsa.subspace_analysis`).
        If specified, it looks `adata_learning.uns[subspace_key]` for subspace settings
        and `adata_learning.varm[adata_learning.uns[subspace_key]['discriminant_key']]` for discriminant matrix.
    copy
        Return a copy instead of writing to ``adata``.
    
    Returns
    -------
    Depending on ``copy``, returns or updates ``adata`` with the following fields.
    
    .obs[``groups``]
        Array of dim (number of samples) that stores the transferred labels for each cell.
    .obsm['transfer_proba'] : :class:`~pandas.DataFrame`
        DataFrame storing the transferred probability for each cell for each group.
    '''
    
    adata = adata.copy() if copy else adata
    
    if 'is_learning' not in adata.obs.keys():
        raise ValueError(
            'Could not find adata.obs[\'is_learning\'].'
            'Please run select_learning_samples first.'
        )
    
    if subspace_key is None:
        subspace_key = 'subspace_analysis'
    
    if subspace_key not in adata_learning.uns.keys():
        raise ValueError(
            f'Could not find adata_learning.uns[\'{subspace_key}\'].'
            'Please run subspace_analysis first.'
        )
    
    if n_discriminant is None:
        n_discriminant = adata_learning.uns[subspace_key]['discriminant'].shape[1]
    
    if groups not in adata_learning.obs.keys():
        raise KeyError(
            f'Could not find key {groups} in adata_learning.obs.columns.'
        )
    
    adata_use = adata[:, adata_learning.uns[subspace_key]['var_names_use']]
    
    adata_learning_use = adata_learning[:, adata_learning.uns[subspace_key]['var_names_use']]
    
    
    clusters_learning = adata_learning_use.obs[groups]
    P = adata_learning.uns[subspace_key]['discriminant'][:, :n_discriminant]
    
    X_learning = adata_learning_use.X.toarray() if issparse(adata_learning_use.X) else adata_learning_use.X
    X_prediction = adata_use[~adata_use.obs['is_learning'], :].X.toarray() if issparse(adata_use.X) else adata_use[~adata_use.obs['is_learning'], :].X
    
    neigh = KNeighborsClassifier(n_neighbors=n_neighbors, metric=metric)
    neigh.fit(np.matmul(X_learning, P), clusters_learning)
    PX = np.matmul(X_prediction, P)
    clusters_prediction = neigh.predict(PX)
    
    clusters = pd.Series(index=adata_use.obs_names, dtype=int)
    clusters.loc[adata_use.obs['is_learning']] = clusters_learning
    clusters.loc[~adata_use.obs['is_learning']] = clusters_prediction
    
    clusters_proba_columns = np.sort(adata_learning.obs[groups].unique())
    clusters_proba_prediction = pd.DataFrame(neigh.predict_proba(PX),
                                             index=adata.obs_names[~adata.obs['is_learning']],
                                             columns=clusters_proba_columns)
    clusters_proba = pd.DataFrame(0., index=adata.obs_names, columns=clusters_proba_columns)
    clusters_proba.loc[adata.obs['is_learning'], :] = pd.get_dummies(clusters_learning, dtype=float)
    clusters_proba.loc[~adata.obs['is_learning'], :] = clusters_proba_prediction
    
    
    adata.uns['transfer_labels'] = {}
    
    transfer_dict = adata.uns['transfer_labels']
    
    transfer_dict['params'] = {}
    transfer_dict['params']['groups'] = groups
    transfer_dict['params']['n_neighbors'] = n_neighbors
    transfer_dict['params']['metric'] = str(metric)
    transfer_dict['params']['n_discriminant'] = n_discriminant
    transfer_dict['params']['subspace_key'] = subspace_key
    
    adata.obsm['transfer_proba'] = clusters_proba
    
    adata.obs[groups] = pd.Categorical(
        values=clusters.astype('U'),
        categories=natsorted(map(str, np.unique(clusters))),
    )
    
    return adata if copy else None






















# -*- coding: utf-8 -*-

from typing import Optional
from anndata import AnnData

import numpy as np
import pandas as pd

from scipy.sparse import issparse
from sklearn.preprocessing import normalize
from tqdm import trange

from ._lasso_landmarks import lasso_landmarks


def select_landmarks(
    adata: AnnData,
    n_landmarks: int,
    Lambda: float = 0.5,
    reltol: float = 1e-3,
    use_highly_variable: Optional[bool] = None,
    random_state: int = 0,
    copy: bool = False,
) -> Optional[AnnData]:
    '''
    Select landmark samples [Matsushima19]_.
    
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_landmarks
        Number of landmarks to be selected.
    Lambda
        Hyperparameter for sparsity regularization.
    reltol
        Relative tolerance in optimization.
    use_highly_variable
        Whether to use highly variable genes only, stored in `adata.var['highly_variable']`.
        By default uses them if they have been determined beforehand.
    random_state
        Change to use different initial states for the optimization.
    copy
        Return a copy instead of writing to ``adata``.
    
    Returns
    -------
    Depending on ``copy``, returns or updates ``adata`` with the following fields.
    
    .obs['is_landmark']
        Boolean indicator of landmark samples.
    '''
    
    adata = adata.copy() if copy else adata
    
    if adata.n_obs < n_landmarks:
        n_landmarks = adata.n_obs
    
    if use_highly_variable is True and 'highly_variable' not in adata.var.keys():
        raise ValueError(
            'Did not find adata.var[\'highly_variable\']. '
            'Either your data already only consists of highly-variable genes '
            'or consider running `pp.highly_variable_genes` first.'
        )
    
    if use_highly_variable is None:
        use_highly_variable = True if 'highly_variable' in adata.var.keys() else False
    
    adata_use = (
        adata[:, adata.var['highly_variable']] if use_highly_variable else adata
    )
    
    
    X = adata_use.X.toarray().T if issparse(adata_use.X) else adata_use.X.T
    X = normalize(X.astype(np.float64), axis=0)
    
    rng = np.random.RandomState(seed=random_state)
    
    n = adata.n_obs
    
    max_t = n_landmarks * 2
    num_I_t = 1
    num_rand_idx = n_landmarks * 2 * num_I_t
    if n >= num_rand_idx:
        rand_idx = rng.choice(np.arange(n), num_rand_idx)
    else:
        rand_idx = []
        for it in range(num_rand_idx // n + 1):
            rand_idx.append(rng.permutation(n))
        rand_idx = np.concatenate(rand_idx)
    
    S_t = np.array([], dtype=int)
    
    norms = np.sum(X**2, axis=0)
    
    stats = {
        'reltol': reltol,
        'Lambda': Lambda,
        'normsSt': np.array([]),
        'XSt': np.ndarray((X.shape[0], 0)),
        'W': np.zeros((n_landmarks, n)),
        'R': X,
    }
    
    # select landmark samples
    pbar = trange(max_t)
    
    for t in pbar:
        
        I_t = rand_idx[(t*num_I_t):((t+1)*num_I_t)]
        
        if S_t.shape[0] != 0:
            for i in I_t:
                stats = lasso_landmarks(X, S_t, stats, i, rng)
        
        candidate = np.setdiff1d(np.arange(n), S_t)
        
        grad_L = np.matmul(stats['R'][:, I_t].T, X[:, candidate])
        grad = np.zeros((num_I_t, n))
        grad[:, candidate] = np.minimum(grad_L+Lambda, np.maximum(0, grad_L-Lambda))
        
        for j in range(num_I_t):
            grad[j, I_t[j]] = 0
        
        grad2sum = np.sum(grad**2, axis=0)
        idx = np.argsort(grad2sum)[::-1]
        max_vals = grad2sum[idx]
        
        if max_vals[0] != 0:
            dSt = idx[0]
            S_t = np.append(S_t, dSt)
            pbar.set_postfix_str(f'seleted landmarks: {S_t.shape[0]}')
            stats['normsSt'] = np.append(stats['normsSt'], norms[dSt])
            stats['XSt'] = np.concatenate((stats['XSt'], X[:, dSt][:, None]), axis=1)
            if S_t.shape[0] == n_landmarks:
                break
    
    landmarks = pd.Series(False, index=adata.obs_names)
    landmarks.iloc[S_t] = True
    
    
    adata.uns['select_landmarks'] = {}
    
    landmarks_dict = adata.uns['select_landmarks']
    
    landmarks_dict['params'] = {}
    landmarks_dict['params']['n_landmarks'] = np.count_nonzero(landmarks)
    landmarks_dict['params']['Lambda'] = Lambda
    landmarks_dict['params']['reltol'] = reltol
    landmarks_dict['params']['use_highly_variable'] = use_highly_variable
    landmarks_dict['params']['random_state'] = random_state
    
    adata.obs['is_landmark'] = landmarks
    
    return adata if copy else None






















# -*- coding: utf-8 -*-

'''
LANdmark-based Transferable Subspace Analysis
'''

__version__ = '0.4.0'


from ._select_learning_samples import select_learning_samples
from ._select_landmarks import select_landmarks
from ._subspace_analysis import subspace_analysis
from ._transfer_labels import transfer_labels












# -*- coding: utf-8 -*-

import torch

from tqdm import trange


def solve_l1l2(W, Lambda):
    
    NW = torch.norm(W, 2, dim=0, keepdim=True)
    E = (NW > Lambda) * (NW - Lambda) / NW
    
    return E * W


def solve_Z(X, Lambda, Zi, n_iterations, device):
    
    mu = 1e-6
    max_mu = 1e30
    rho = 1.1
    tol1 = 1e-4
    tol2 = 1e-5
    
    m, n = X.shape
    n = torch.tensor(n).to(device)
    normfX = torch.norm(X, 'fro')
    norm2X = torch.norm(X, 2)
    eta = norm2X ** 2 + n + 1
    tensor0 = torch.tensor(0).to(device)
    
    # intialize
    J = torch.zeros(n, n).to(device)
    Z = torch.zeros(n, n).to(device)
    E = torch.zeros(m, n).to(device)
    
    Y1 = torch.zeros(m, n).to(device)
    Y2 = torch.zeros(1, n).to(device)
    Y3 = torch.zeros(n, n).to(device)
    
    pbar = trange(n_iterations)
    
    for Iter in pbar:
        Em = E
        Zm = Z
        
        temp = Z + Y3 / mu
        
        J = torch.maximum(temp - 1 / mu, tensor0)
        
        temp = X - torch.matmul(X, Z) + Y1 / mu
        E = solve_l1l2(temp, Lambda / mu)
        
        H = - torch.matmul(X.T, (X - torch.matmul(X, Z) - E + Y1 / mu)) - (1 - torch.sum(Z, axis=0, keepdims=True) + Y2 / mu) + (Z - J + Y3 / mu)
        M = Z - H / eta
        Z = Zi * M
        
        xmaz = X - torch.matmul(X, Z)
        leq1 = xmaz - E
        leq2 = 1 - torch.sum(Z, axis=0, keepdims=True)
        leq3 = Z - J
        relChgZ = torch.norm(Z - Zm, 'fro') / normfX
        relChgE = torch.norm(E - Em, 'fro') / normfX
        relChg = torch.max(relChgE, relChgZ)
        recErr = torch.norm(leq1, 'fro') / normfX
        
        pbar.set_postfix_str(f'relChg: {relChg.item():.3e}, recErr: {recErr.item():.3e}')
        
        if relChg < tol1 and recErr < tol2:
            pbar.set_postfix_str(f'relChg: {relChg.item():.3e}, recErr: {recErr.item():.3e}, converged!')
            break
        else:
            Y1 = Y1 + mu * leq1
            Y2 = Y2 + mu * leq2
            Y3 = Y3 + mu * leq3
            mu = min(max_mu, mu * rho)
    
    return Z, E


def solve_P(X, Z, P):
    
    eps = 1e-14
    temp = torch.matmul(P.T, X - torch.matmul(X, Z))
    D = torch.diag(0.5 / torch.norm(temp, 2, dim=0) + eps)
    S = torch.matmul(torch.matmul(X - torch.matmul(X, Z), D), (X - torch.matmul(X, Z)).T)
    S = (S + S.T) / 2
    DS, Pall = torch.eig(S, eigenvectors=True)
    Pu = Pall[:, DS[:, 0] > 10^-3]
    
    return Pu






















# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lantsa']

package_data = \
{'': ['*']}

install_requires = \
['anndata>=0.7.5',
 'h5py>=2.9.0',
 'numba>=0.41.0',
 'numpy>=1.17.0',
 'pandas>=1.0',
 'scikit-learn>=0.21.2',
 'torch>=1.7.0',
 'tqdm>=4.56.0']

extras_require = \
{'docs': ['nbsphinx',
          'pydata-sphinx-theme>=0.4.3',
          'scanpydoc>=0.5',
          'sphinx>=3.4',
          'sphinx-autodoc-typehints',
          'sphinx_copybutton<=0.3.1'],
 'docs:python_version >= "3.7"': ['ipython>=7.20'],
 'tutorials': ['leidenalg',
               'python-igraph',
               'scanpy>=1.6',
               'scikit-misc>=0.1.3']}

setup_kwargs = {
    'name': 'lantsa',
    'version': '0.4.0',
    'description': 'Landmark-based transferable subspace analysis for single-cell and spatial transcriptomics',
    'long_description': '# LANTSA\n\nLANTSA is a landmark-based transferable subspace analysis tool for single-cell and spatial transcriptomics.\n\nPlease visit [LANTSA documentation website](https://lantsa.readthedocs.io/) for details about installation, tutorials, API and references.\n',
    'author': 'Lequn Wang',
    'author_email': 'wanglequn2019@sibcb.ac.cn',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

from typing import Optional
from anndata import AnnData
from ._compat import Literal

import numpy as np
import pandas as pd

from scipy.sparse import issparse
from sklearn.decomposition import PCA
from sklearn.cluster import MiniBatchKMeans
from sklearn.metrics.pairwise import euclidean_distances

_Method = Literal['MiniBatchKMeans', 'Random']


def select_learning_samples(
    adata: AnnData,
    n_learning: int,
    method: _Method = 'MiniBatchKMeans',
    n_components: Optional[int] = None,
    use_highly_variable: Optional[bool] = None,
    random_state: int = 0,
    copy: bool = False,
) -> Optional[AnnData]:
    '''
    Select learning samples.
    
    Parameters
    ----------
    adata
        Annotated data matrix.
    n_learning
        Number of learning samples to be selected.
    method
        Method to use for learning sample selection.
        
        * ``'MiniBatchKMeans'``
            Use `scikit-learn` :class:`~sklearn.cluster.MiniBatchKMeans`
            to select learning samples.
        * ``'Random'``
            Randomly choose learning samples.
        
    n_components
        Number of principal components to use as the input of
        :class:`~sklearn.cluster.MiniBatchKMeans`.
        Defaults to the number of features devided by 50 with a minimum number of 20.
    use_highly_variable
        Whether to use highly variable genes only, stored in `adata.var['highly_variable']`.
        By default uses them if they have been determined beforehand.
    random_state
        Change to use different initial states for the optimization.
    copy
        Return a copy instead of writing to ``adata``.
    
    Returns
    -------
    Depending on ``copy``, returns or updates ``adata`` with the following fields.
    
    .obs['is_learning']
        Boolean indicator of learning samples.
    '''
    
    adata = adata.copy() if copy else adata
    
    if method not in ['MiniBatchKMeans', 'Random']:
        raise ValueError('method needs to be \'MiniBatchKMeans\' or \'Random\'')
    
    if n_learning > adata.n_obs:
        raise ValueError(f'Expected n_learning <= n_obs, but n_learning = {n_learning}, n_obs = {adata.shape[0]}')
    
    if use_highly_variable is True and 'highly_variable' not in adata.var.keys():
        raise ValueError(
            'Did not find adata.var[\'highly_variable\']. '
            'Either your data already only consists of highly-variable genes '
            'or consider running `pp.highly_variable_genes` first.'
        )
    
    if use_highly_variable is None:
        use_highly_variable = True if 'highly_variable' in adata.var.keys() else False
    
    adata_use = (
        adata[:, adata.var['highly_variable']] if use_highly_variable else adata
    )
    
    
    if method == 'MiniBatchKMeans':
        
        X = adata_use.X.toarray() if issparse(adata_use.X) else adata_use.X
        
        if n_components is None:
            n_components = min(max(adata_use.n_vars // 50, 20), adata.n_obs-1)
        if n_components < adata_use.n_vars:
            X = PCA(n_components=n_components, svd_solver='arpack', random_state=random_state).fit_transform(X)
        else:
            n_components = None
        
        kmeans = MiniBatchKMeans(n_clusters=n_learning, random_state=random_state).fit(X)
        learning_samples = pd.Series(False, index=adata_use.obs_names)
        learning_samples.iloc[np.unique(np.argmin(euclidean_distances(X, kmeans.cluster_centers_), axis=0))] = True
    
    elif method == 'Random':
        
        n_components = None
        
        rng = np.random.RandomState(seed=random_state)
        learning_samples = pd.Series(False, index=adata_use.obs_names)
        learning_samples.iloc[rng.choice(np.arange(adata_use.shape[0]), size=n_learning, replace=False)] = True
    
    
    adata.uns['select_training_samples'] = {}
    
    learn_dict = adata.uns['select_training_samples']
    
    learn_dict['params'] = {}
    learn_dict['params']['n_learning'] = np.count_nonzero(learning_samples)
    learn_dict['params']['method'] = method
    learn_dict['params']['n_components'] = n_components
    learn_dict['params']['use_highly_variable'] = use_highly_variable
    learn_dict['params']['random_state'] = random_state
    
    adata.obs['is_learning'] = learning_samples
    
    return adata if copy else None



















